/*  Program for Informative Sites Test - PIST, version 1.0 */
/*  (c) Copyright 2001, Andrew Rambaut & Michael Worobey   */
/*  Department of Zoology, University of Oxford            */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "tree.h"
#include "evolve.h"
#include "models.h"
#include "gamma.h"
#include "random.h"

int numTaxa, numBases, fileFormat;

double gammaShape, propInvariant;
int numCats, rateHetero, invariantSites;
double catRate[MAX_RATE_CATS];
double freqRate[MAX_RATE_CATS];

char *nucleotides="ACGT";
static double matrix[MAX_RATE_CATS][16];
static double vector[4];
static double *siteRates;
static short *categories;
static short *invariant;

/* prototypes */

char SetBase(double *P);
short IsInvariant();
void SetSequence(char *seq, char *source);
void RandomSequence(char *seq);
void MutateSequence(char *seq, double len);
void EvolveNode(TNode *anc, TNode *des, double scale);

void WritePhylipFormat(FILE *fv, TTree *tree);
void WriteNexusFormat(FILE *fv, TTree *tree);

/* functions */

void CreateSequences(TTree *tree)
{
	TNode *P;
	
	P=tree->nodeList;
	while (P!=NULL) {
		P->sequence=CAllocMem(numBases+1, "sequences", "CreateSequences", 0);
		P=P->next;
	}
	if (rateHetero==GammaRates)
		siteRates=CAllocMem(numBases*sizeof(double), "siteRates", "CreateSequences", 0);
	else if (rateHetero==DiscreteGammaRates)
		categories=CAllocMem(numBases*sizeof(short), "categories", "CreateSequences", 0);
		
	if (invariantSites)
		invariant=CAllocMem(numBases*sizeof(short), "invariant", "CreateSequences", 0);
}

void SetCategories()
{
	int i;
	double sumRates;
	
	if (rateHetero==CodonRates) {
		sumRates=catRate[0]+catRate[1]+catRate[2];
		if (sumRates!=3.0) {
			catRate[0]*=3.0/sumRates;
			catRate[1]*=3.0/sumRates;
			catRate[2]*=3.0/sumRates;
		}
	} else if (rateHetero==GammaRates) {
		for (i=0; i<numBases; i++)
			siteRates[i]=rndgamma(gammaShape) / gammaShape;

	} else if (rateHetero==DiscreteGammaRates) {
		DiscreteGamma(freqRate, catRate, gammaShape, gammaShape, numCats, 0);
		for (i=0; i<numBases; i++)
			categories[i]=(int)(rndu()*numCats);
	}
	
	if (invariantSites) {
		for (i=0; i<numBases; i++)
			invariant[i] = IsInvariant();
	} 
}


char SetBase(double *P)
{
	char j;
	double r;
	
	r=rndu();
	for (j='\0'; r>(*P) && j<'\3'; j++) P++;
	return j;
}


short IsInvariant()
{
	double r;

	r=rndu();
	if (r < propInvariant)
		return 1;
	else 
		return 0;
}


void RandomSequence(char *seq)
{
	int i;
	char *P;
	
	P=seq;
	for (i=0; i<numBases; i++) {
		*P=SetBase(addFreq);
		P++;
	}
}

void SetSequence(char *seq, char *source)
{
	int i;
	char *P, *Q;
	
	P=seq;
	Q=source;
	for (i=0; i<numBases; i++) {
		switch (*Q) {
			case 'A': *P='\0'; break;
			case 'C': *P='\1'; break;
			case 'G': *P='\2'; break;
			case 'T': *P='\3'; break;
			default:
				fprintf(stderr, "Bad base in ancestoral sequence\n");
				exit(0);
		}
		P++;
		Q++;
	}
}

void MutateSequence(char *seq, double len)
{
	int i, cat;
	double *Q;
	short *R;
	short *S;
	char *P;
	
	P=seq;
	
	switch (rateHetero) {
		case GammaRates:
			Q=siteRates;
			
			if (invariantSites) {
				S = invariant;
				for (i=0; i<numBases; i++) {
					if (!(*S)) {
						SetVector(vector, *P, (*Q)*len);
						*P=SetBase(vector);
					}
					P++;
					Q++;
					S++;
				}
			} else {
				for (i=0; i<numBases; i++) {
					SetVector(vector, *P, (*Q)*len);
					*P=SetBase(vector);
					P++;
					Q++;
				}
			}
		break;
		case DiscreteGammaRates:
			for (i=0; i<numCats; i++)
				SetMatrix(matrix[i], catRate[i]*len);
			
			R=categories;
			if (invariantSites) {
				S = invariant;
				for (i=0; i<numBases; i++) {
					if (!(*S))
						*P=SetBase(matrix[*R]+(*P<<2));
					P++;
					R++;
					S++;
				}
			} else {
				for (i=0; i<numBases; i++) {
					*P=SetBase(matrix[*R]+(*P<<2));
					P++;
					R++;
				}
			}
		break;
		case CodonRates:
			for (i=0; i<numCats; i++)
				SetMatrix(matrix[i], catRate[i]*len);
			
			for (i=0; i<numBases; i++) {
				cat=i%3;
				*P=SetBase(matrix[cat]+(*P<<2));
				P++;
			}
		break;
		case NoRates:
			SetMatrix(matrix[0], len);
			
			if (invariantSites) {
				S = invariant;
				for (i=0; i<numBases; i++) {
					if (!(*S))
						*P=SetBase(matrix[0]+(*P<<2));
					P++;
					S++;
				}
			} else {
				for (i=0; i<numBases; i++) {
					*P=SetBase(matrix[0]+(*P<<2));
					P++;
				}
			}
		break;
	}
}


void EvolveNode(TNode *anc, TNode *des, double scale)
{
	double len;
	
	if (scale<0.0)
		len=des->length0;
	else
		len=des->length0*scale;
		
	memcpy(des->sequence, anc->sequence, numBases);
	MutateSequence(des->sequence, len);
	
	if (des->tipNo==-1) {
		EvolveNode(des, des->branch1, scale);
		EvolveNode(des, des->branch2, scale);
	}
}


void EvolveSequences(TTree *tree, double scale, char *ancestor)
{
	if (ancestor==NULL)
		RandomSequence(tree->root->sequence);
	else
		SetSequence(tree->root->sequence, ancestor);
	
	/* Rescale the branch lengths to accommodate the invariant sites */
	if (invariantSites)
		scale /= (1.0 - propInvariant);

	EvolveNode(tree->root, tree->root->branch1, scale);
	EvolveNode(tree->root, tree->root->branch2, scale);
	if (!tree->rooted)
		EvolveNode(tree->root, tree->root->branch0, scale);
}


void WriteSequences(FILE *fv, TTree *tree)
{
	switch (fileFormat) {
		case PHYLIPFormat:
		case RelaxedFormat:
			WritePhylipFormat(fv, tree);
		break;
		case NEXUSFormat:
			WriteNexusFormat(fv, tree);
		break;
	}
}

void WritePhylipFormat(FILE *fv, TTree *tree)
{
	int i, j;
	char *P;

	fprintf(fv, " %d %d\n", numTaxa, numBases);

	for (i=0; i<numTaxa; i++) {
		if (fileFormat == RelaxedFormat)
			fprintf(fv, "%s ", tree->names[i]);
		else {
			j=0;
			P=tree->names[i];
			while (j<10 && *P) {
				fputc(*P, fv);
				j++;
				P++;
			}
			while (j<10) {
				fputc(' ', fv);
				j++;
			}
		}
		P=tree->tips[i]->sequence;
		for (j=0; j<numBases; j++) {
			fputc(nucleotides[*P], fv);
			P++;
		}
		fputc('\n', fv);
	}
}

void WriteNexusFormat(FILE *fv, TTree *tree)
{
	int i, j, len, maxLen;
	char *P;

	fprintf(fv, "Begin DATA;\n");
	fprintf(fv, "\tDimensions NTAX=%d NCHAR=%d;\n", numTaxa, numBases);
	fprintf(fv, "\tFormat DATATYPE=DNA MISSING=? GAP=-;\n");
	fprintf(fv, "\tMatrix\n");

	maxLen = strlen(tree->names[0]);
	for (i=1; i<numTaxa; i++) {
		len = strlen(tree->names[i]);
		if (len > maxLen)
			maxLen = len;
	}

	for (i=0; i<numTaxa; i++) {
		fprintf(fv, "%s ", tree->names[i]);
		len = maxLen - strlen(tree->names[i]);
		for (j = 0; j < len; j++)
			fputc(' ', fv);

		P=tree->tips[i]->sequence;
		for (j=0; j<numBases; j++) {
			fputc(nucleotides[*P], fv);
			P++;
		}
		fputc('\n', fv);
	}
	fprintf(fv, "\t;\nEND;\n\n");
}

