/*  Parsimony tree length calculator                       */
/*  Program for Informative Sites Test - PIST, version 1.0 */
/*  (c) Copyright 2001, Andrew Rambaut & Michael Worobey   */
/*  Department of Zoology, University of Oxford            */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "parsimony.h"
#include "evolve.h"
#include "tree.h"

#define MAX_NUM_STATES 4

/* prototypes */

void CalcNodeLength(TNode *node, int site, char *outSet, long *outScore, int numStates);
int IsEmptySet(char *inSet, int numStates);
void IntersectSets(char *inSet1, char *inSet2, char *outSet, int numStates);
void UnionSets(char *inSet1, char *inSet2, char *outSet, int numStates);


/* functions */

long GetNumInformativeSites(char **sequences, long *outNumPolymorphicSites, int twoStates)
{
	long numPolySites = 0;
	long numInfSites = 0;
	int numStates;
	int count[MAX_NUM_STATES];
	
	long i, j, k;
	for (i=0; i<numBases; i++) {
		count[0] = 0;
		count[1] = 0;
		count[2] = 0;
		count[3] = 0;
		
		for (j=0; j<numTaxa; j++) {
			switch (sequences[j][i]) {
				case 'A': case '\0': count[0]++; break;
				case 'C': case '\1': count[1]++; break;
				case 'G': case '\2': count[2]++; break;
				case 'T': case '\3': count[3]++; break;
			}
		}
		
		numStates = 0;
		if (count[0]>0)
			numStates++;
		if (count[1]>0)
			numStates++;	
		if (count[2]>0)
			numStates++;	
		if (count[3]>0)
			numStates++;	
			
		if (twoStates == 0 || numStates == 2) {
			k = 0;
			if (count[0]>1)
				k++;
			if (count[1]>1)
				k++;	
			if (count[2]>1)
				k++;	
			if (count[3]>1)
				k++;	
			
			if (k > 1)
				numInfSites++;
		}
		
		k = 0;
		if (count[0]==0)
			k++;
		if (count[1]==0)
			k++;	
		if (count[2]==0)
			k++;	
		if (count[3]==0)
			k++;	
		
		if (k < 3)
			numPolySites++;
	}
		
	if (outNumPolymorphicSites != NULL)
		*outNumPolymorphicSites = numPolySites;
		
	return numInfSites;
}

long GetNumInformativeSitesOnTree(TTree *tree, long *outNumPolymorphicSites, int twoStates)
{
	long numPolySites = 0;
	long numInfSites = 0;
	int numStates;
	int count[MAX_NUM_STATES];
	
	long i, j, k;
	for (i=0; i<numBases; i++) {
		count[0] = 0;
		count[1] = 0;
		count[2] = 0;
		count[3] = 0;
		
		for (j=0; j<numTaxa; j++) {
			switch (tree->tips[j]->sequence[i]) {
				case 'A': case '\0': count[0]++; break;
				case 'C': case '\1': count[1]++; break;
				case 'G': case '\2': count[2]++; break;
				case 'T': case '\3': count[3]++; break;
			}
		}
		
		numStates = 0;
		if (count[0]>0)
			numStates++;
		if (count[1]>0)
			numStates++;	
		if (count[2]>0)
			numStates++;	
		if (count[3]>0)
			numStates++;	
			
		if (twoStates == 0 || numStates == 2) {
			k = 0;
			if (count[0]>1)
				k++;
			if (count[1]>1)
				k++;	
			if (count[2]>1)
				k++;	
			if (count[3]>1)
				k++;	
			
			if (k > 1)
				numInfSites++;
		}
		
		k = 0;
		if (count[0]==0)
			k++;
		if (count[1]==0)
			k++;	
		if (count[2]==0)
			k++;	
		if (count[3]==0)
			k++;	
		
		if (k < 3)
			numPolySites++;
	}
		
	if (outNumPolymorphicSites != NULL)
		*outNumPolymorphicSites = numPolySites;
		
	return numInfSites;
}

long CalcTreeLength(TTree *tree)
{
	long score = 0;
	long siteScore;
	
	char set1[MAX_NUM_STATES];
	char set2[MAX_NUM_STATES];

	int i;
	for (i=0; i<numBases; i++) {
		siteScore = 0;
		if (!tree->rooted) {
			CalcNodeLength(tree->root, i, set1, &siteScore, 4);
			CalcNodeLength(tree->root->branch0, i, set2, &siteScore, 4);
		} else {
			CalcNodeLength(tree->root->branch1, i, set1, &siteScore, 4);
			CalcNodeLength(tree->root->branch2, i, set2, &siteScore, 4);
		}
				
		IntersectSets(set1, set2, set1, 4);		
		if (IsEmptySet(set1, 4))
			siteScore++;
					
		score += siteScore;
	}
	
	return score;
}

void CalcNodeLength(TNode *node, int site, char *outSet, long *outScore, int numStates)
{
	char iSet[MAX_NUM_STATES];
	char uSet[MAX_NUM_STATES];
	char set[MAX_NUM_STATES];
	
	if (node->tipNo==-1) {		
		CalcNodeLength(node->branch1, site, iSet, outScore, numStates);
		memcpy(uSet, iSet, numStates);
		
		CalcNodeLength(node->branch2, site, set, outScore, numStates);
		IntersectSets(iSet, set, iSet, numStates);
		UnionSets(uSet, set, uSet, numStates);
		
		if (IsEmptySet(iSet, numStates)) {
			(*outScore)++;
			memcpy(outSet, uSet, numStates);
		} else
			memcpy(outSet, iSet, numStates);
	} else {
		outSet[0] = 0;
		outSet[1] = 0;
		outSet[2] = 0;
		outSet[3] = 0;
		switch (node->sequence[site]) {
			case 'A': case '\0': outSet[0] = 1; break;
			case 'C': case '\1': outSet[1] = 1; break;
			case 'G': case '\2': outSet[2] = 1; break;
			case 'T': case '\3': outSet[3] = 1; break;
			default:
				outSet[0] = 1;
				outSet[1] = 1;
				outSet[2] = 1;
				outSet[3] = 1;
			break;
		}
	}
}

int IsEmptySet(char *inSet, int numStates)
{
	int i;
	for (i=0; i<numStates; i++) {
		if ( inSet[i]==1 )
			return 0;
	}
	
	return 1;
}

void IntersectSets(char *inSet1, char *inSet2, char *outSet, int numStates)
{
	int i;
	for (i=0; i<numStates; i++) {
		if ( inSet1[i] && inSet2[i] )
			outSet[i]=1;
		else
			outSet[i]=0;
	}
}

void UnionSets(char *inSet1, char *inSet2, char *outSet, int numStates)
{
	int i;
	for (i=0; i<numStates; i++) {
		if ( inSet1[i] || inSet2[i] )
			outSet[i]=1;
		else
			outSet[i]=0;
	}
}
