/*  Program for Informative Sites Test - PIST, version 1.0 */
/*  (c) Copyright 2001, Andrew Rambaut & Michael Worobey   */
/*  Department of Zoology, University of Oxford            */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#ifdef __MWERKS__
#include <console.h>
#endif

#include "global.h"
#include "treefile.h"
#include "evolve.h"
#include "parsimony.h"
#include "models.h"
#include "progress.h"
#include "random.h"

#define PROGRAM_NAME "pist"
#define VERSION_NUMBER "Version 1.0"

char *ModelNames[numModels]={
	"F84",
	"HKY",
	"REV"
};

int printHelp, treeFile, numDatasets, numTrees;
int statsOnly, writeSeqs, hasRandomLength, twoStates;
double acceptInterval, randomLength;
char treeFileName[256];

int hasAlignment, numSequences, numSites;

char **names;
char **sequences;

FILE *tree_fv;

/* prototypes */

static void PrintTitle();
static void PrintUsage();
static void PrintVerbose();
static void ReadParams();
static void ReadFileParams();
static void AllocateMemory();
static void ReadFile();
static int OpenTreeFile();
static void SetTipSequences(TTree *tree);
static void WriteStats();
static int Compare(const double *val1, const double *val2);
 
/* functions */

static void PrintTitle()
{
	fprintf(stderr, "Program for the Informative Sites Test - %s\n", PROGRAM_NAME);
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 2001 Andrew Rambaut and Michael Worobey\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{ 
	fprintf(stderr, "Usage: pist [-m MODEL] [-n #] [-p #] [-l #] [-c #1 #2 #3 | [-a #] [-g #] [-i #]]\n");
	fprintf(stderr, "            [-f #A #C #G #T] [-t # | -r #1 #2 #3 #4 #5 # 6] [-k #] [-p]\n");
	fprintf(stderr, "            [-s] [-w] [-o[p][r][n]] [-q] [-h] [treefile]\n");
	fprintf(stderr, "  -m: MODEL = F84, HKY, REV [default = HKY]\n");
	fprintf(stderr, "  -n: # = simulated acceptable datasets [default = 1000].\n");
	fprintf(stderr, "  -p: # = tree length acceptance interval (%) [default = 0.5].\n");
	fprintf(stderr, "  -l: # = average tree length for randomised data [default = none].\n");
/*	fprintf(stderr, "  -2: Use only 2-state informative sites [default = all].\n");*/
	fprintf(stderr, "  -c: #1 #2 #3 = rates for codon position heterogeneity [default = none].\n");
	fprintf(stderr, "  -a: # = shape (alpha) for gamma rate heterogeneity [default = none].\n");
	fprintf(stderr, "  -g: # = number of gamma rate categories [default = continuous].\n");
	fprintf(stderr, "  -i: # = proportion of invariable sites [default = 0.0].\n");
	fprintf(stderr, "  -t: # = transition-transversion ratio [default = 2.0].\n");
	fprintf(stderr, "  -r: #1 #2 #3 #4 #5 #6= general rate matrix [default = all 1.0].\n");
	fprintf(stderr, "  -f: #A #C #G #T = nucleotide frequencies [default = all equal].\n");
	fprintf(stderr, "  -s: Write statistics only (can take multiple input datasets)\n");
	fprintf(stderr, "  -w: Write sequences to file 'pist.seq'\n");
	fprintf(stderr, "  -o: Output file format [default = PHYLIP]\n");
	fprintf(stderr, "    p PHYLIP format\n");
	fprintf(stderr, "    r relaxed PHYLIP format\n");
	fprintf(stderr, "    n NEXUS format\n");
	fprintf(stderr, "  -h: Give this help message\n");
	fprintf(stderr, "  -q: Quiet\n");
	fprintf(stderr, "  treefile: name of tree file [default = trees on stdin]\n\n");
}


void ReadParams(int argc, char **argv)
{
	int i, j;
	char ch, *P, st[4];
	
	model=HKY;
	
	numCats=1;
	rateHetero=NoRates;
	catRate[0]=1.0;
	gammaShape=1.0;
	invariantSites=0;
	propInvariant=0.0;
		
	freq[A]=freq[C]=freq[G]=freq[T]=0.25;
	tstv=2.0;
	
	Rmat[0]=Rmat[1]=Rmat[2]=Rmat[3]=Rmat[4]=1.0;
	
	numBases=-1;
	numDatasets=100;
	
 	verbose=1;
	relaxedPhylip=1;
	quiet=0;
	
	statsOnly = 0;
	writeSeqs = 0;
	fileFormat = PHYLIPFormat;
	acceptInterval = 0.5;
	hasRandomLength = 0;
	twoStates = 1;
	
	treeFile=0;
		
	for (i=1; i<argc; i++) {
		P=argv[i];
		if (*P!='-') {
			if (treeFile) {
				fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
				PrintUsage();
				exit(0);
			}
			treeFile=1;
			strcpy(treeFileName, argv[i]);
		} else {
			while (*P) {
				*P=toupper(*P);
				P++;
			}
			P=argv[i];
			P++;
			ch=*P;
			P++;
			switch (ch) {
				case 'H':
					printHelp=1;
				break;
				case 'M':
					if (GetStrParam(argc, argv, &i, P, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						exit(0);
					}
					
					P=st;
					model=-1;
					for (j=F84; j<numModels; j++) {
						if (strncmp(P, ModelNames[j], 3)==0)
							model=j;
					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						exit(0);
					}

				break;
				case 'N':
					if (GetIntParams(argc, argv, &i, P, 1, &numDatasets) || numDatasets<1) {
						fprintf(stderr, "Bad (or missing) number of datasets: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'P':
					if (GetDoubleParams(argc, argv, &i, P, 1, &acceptInterval)) {
						fprintf(stderr, "Bad Acceptence Interval: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'L':
					hasRandomLength = 1;
					if (GetDoubleParams(argc, argv, &i, P, 1, &randomLength)) {
						fprintf(stderr, "Bad Average Randomised Tree Length: %s\n\n", argv[i]);
						exit(0);
					}
				break;
/*				case '2':
					twoStates=1;
				break;*/
				case 'C':
					if (rateHetero==GammaRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates not both\n\n", argv[i]);
						exit(0);
					}
					numCats=3;
					rateHetero=CodonRates;
					if (GetDoubleParams(argc, argv, &i, P, 3, catRate)) {
						fprintf(stderr, "Bad Category Rates: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'A':
					if (rateHetero==CodonRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates not both\n\n", argv[i]);
						exit(0);
					}
					
					if (rateHetero==NoRates)
						rateHetero=GammaRates;
					if (GetDoubleParams(argc, argv, &i, P, 1, &gammaShape) || gammaShape<=0.0) {
						fprintf(stderr, "Bad Gamma Shape: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'G':
					if (rateHetero==CodonRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates/invariant sites not both\n\n", argv[i]);
						exit(0);
					}
					
					rateHetero=DiscreteGammaRates;
					if (GetIntParams(argc, argv, &i, P, 1, &numCats) || numCats<2 || numCats>MAX_RATE_CATS) {
						fprintf(stderr, "Bad number of Gamma Categories: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'I':
					if (rateHetero==CodonRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates/invariant sites not both\n\n", argv[i]);
						exit(0);
					}
					
					invariantSites=1;
					if (GetDoubleParams(argc, argv, &i, P, 1, &propInvariant) || 
										propInvariant<0.0 || propInvariant>=1.0) {
						fprintf(stderr, "Bad Proportion of Invariant Sites: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'F':
					if (GetDoubleParams(argc, argv, &i, P, 4, freq)) {
						fprintf(stderr, "Bad Nucleotide Frequencies: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'T':
					if (GetDoubleParams(argc, argv, &i, P, 1, &tstv)) {
						fprintf(stderr, "Bad Transition-Transversion Ratio: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'R':
					if (GetDoubleParams(argc, argv, &i, P, 6, Rmat)) {
						fprintf(stderr, "Bad General Rate Matrix: %s\n\n", argv[i]);
						exit(0);
					}
					if (Rmat[5]!=1.0) {
						for (j=0; j<5; j++) 
							Rmat[j]/=Rmat[5];
						Rmat[5]=1.0;
					}
				break;
				case 'S':
					statsOnly=1;
				break;
				case 'W':
					writeSeqs=1;
				break;
				case 'O':
					switch (*P) {
						case 'P': fileFormat=PHYLIPFormat; break;
						case 'R': fileFormat=RelaxedFormat; break;
						case 'N': fileFormat=NEXUSFormat; break;
						default:					
							fprintf(stderr, "Unknown output format: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
					}
					writeSeqs=1;
				break;
				case 'Q':
					quiet=1;
				break;
				default:
					fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
					PrintUsage();
					exit(0);
				break;
			}
		}
	}
}

void PrintVerbose()
{
	fprintf(stderr, "Dataset has %d taxa, %d bases\n", numTaxa, numBases);
	fprintf(stderr, "  Null distribution simulated with %d replicates\n\n", numDatasets);
	
	if (rateHetero==CodonRates) {
		fprintf(stderr, "Codon position rate heterogeneity:\n");
		fprintf(stderr, "    rates = 1:%f 2:%f 3:%f\n", catRate[0], catRate[1], catRate[2]);
	} else if (rateHetero==GammaRates) {
		fprintf(stderr, "Continuous gamma rate heterogeneity:\n");
		fprintf(stderr, "    shape = %f\n", gammaShape);
	} else if (rateHetero==DiscreteGammaRates) {
		fprintf(stderr, "Discrete gamma rate heterogeneity:\n");
		fprintf(stderr, "    shape = %f, %d categories\n", gammaShape, numCats);
	} else
		fprintf(stderr, "Rate homogeneity of sites.\n");
	if (invariantSites) {
		fprintf(stderr, "Invariant sites model:\n");
		fprintf(stderr, "    proportion invariant = %f\n", propInvariant);
	}
	fprintf(stderr, "Model=%s\n", ModelNames[model]);
	if (model==F84)
		fprintf(stderr, "  transition/transversion ratio = %G (K=%G)\n", tstv, kappa);
	else if (model==HKY)
		fprintf(stderr, "  transition/transversion ratio = %G (kappa=%G)\n", tstv, kappa);
	else if (model==REV) {
		fprintf(stderr, "  rate matrix = gamma1:%7.4f alpha1:%7.4f  beta1:%7.4f\n", Rmat[0], Rmat[1], Rmat[2]);
		fprintf(stderr, "                                beta2:%7.4f alpha2:%7.4f\n", Rmat[3], Rmat[4]);
		fprintf(stderr, "                                              gamma2: %7.4f\n", Rmat[5]);
/*		fprintf(stderr, "  mean transition/transversion ratio = %G\n", tstv);*/
	}
	fprintf(stderr, "  frequencies = A:%G C:%G G:%G T:%G\n\n", freq[A], freq[C], freq[G], freq[T]);
}

void ReadFileParams()
{
	char ch, st[256];
	
	hasAlignment=0;
	
	ch=fgetc(stdin);
	while (!feof(stdin) && isspace(ch)) 
		ch=fgetc(stdin);
		
	ungetc(ch, stdin);

	if (ch!='(' && isdigit(ch)) {
		fgets(st, 255, stdin);
		if ( sscanf( st, " %d %d", &numSequences, &numSites)!=2 ) {
			fprintf(stderr, "Unable to read parameters from standard input\n");
			exit(0);
		}
		hasAlignment=1;
	}
		
}

void AllocateMemory()
{
	int i;
	
	names=(char **)AllocMem(sizeof(char *)*numSequences, "names", "AllocateMemory", 0);
	sequences=(char **)AllocMem(sizeof(char *)*numSequences, "sequences", "AllocateMemory", 0);
	for (i=0; i<numSequences; i++) {
		names[i]=(char *)AllocMem(sizeof(char)*(MAX_NAME_LEN+1), "names[]", "AllocateMemory", 0);
		sequences[i]=(char *)AllocMem(sizeof(char)*(numSites+1), "sequences[]", "AllocateMemory", 0);
	}
}


void ReadFile()
{
	int n, b, i;
	char ch;
		
	n=0;
	do {
		ch=fgetc(stdin);
		while ( !feof(stdin) && isspace(ch)) 
			ch=fgetc(stdin);
			
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n"); 
			exit(0);
		}
	
		i=0;
		while ( i<MAX_NAME_LEN && !feof(stdin) && !isspace(ch) ) {
			names[n][i]=ch;
			ch=fgetc(stdin);
			i++;
		}
		names[n][i]='\0';
		if (i==0) {
			fprintf(stderr, "Name missing for species %d\n", n+1);
			exit(0);
		}
		while (!feof(stdin) && isspace(ch))
			ch=fgetc(stdin);
		
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		b=0;
		while ( !feof(stdin) && b<numSites) {
			if ( !isspace(ch) ) {
				sequences[n][b]=ch;
				b++;
			}
			ch=toupper(fgetc(stdin));
		}
		
		if ( b<numSites ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		/*fprintf(stderr, "%d: %s, bases read: %d\n", n+1, names[n], b);*/
		n++;
		
		if ( n<numSequences && feof(stdin) ) {
			fprintf(stderr, "Too few sequences in input file\n");
			exit(0);
		}
	} while ( n<numSequences );
}

int OpenTreeFile()
{
	char st[256];
	int n;
		
	if (treeFile) {
		if ( (tree_fv=fopen(treeFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening tree file: '%s'\n", treeFileName);
			exit(0);
		}
		n=CountTrees(tree_fv);
	} else {
		tree_fv=stdin;
		if (hasAlignment) {
			fgets(st, 255, stdin);
			if ( sscanf(st, " %d ", &n)!=1 ) {
				fprintf(stderr, "Tree is missing from end of sequence file\n");
				exit(0);
			}
		} else
			n=CountTrees(stdin);
	}
	
	return n;
}

void SetTipSequences(TTree *tree)
{
	int i, j, n = tree->numTips;
	
	for (i = 0; i < n; i++) {
		j = 0;
		while (j < n && strcmp(names[i],tree->names[j]) != 0)
			j++;
			
		if (j == n) {
			fprintf(stderr, "Taxon name missing from tree: %s\n", names[i]);
			exit(0);
		}
		memcpy(tree->tips[j]->sequence, sequences[i], numBases);
	}
}

void WriteStats()
{
	long realInfSites, realPolySites;
	double observedAlpha;

	AllocateMemory();	
	ReadFile();
	
	numTaxa = numSequences;
	
	realInfSites = GetNumInformativeSites(sequences, &realPolySites, twoStates);
/*	observedAlpha = GetObservedAlpha(sequences);*/

	fprintf(stdout, "NumInformSites\tNumPolySites\n");
	fprintf(stdout, "%d\t%d\n", realInfSites, realPolySites);
}

int Compare(const double *val1, const double *val2)
{
	if (*val1 > *val2) 
		return 1;
	else if (*val1 < *val2)
		return -1;
	else
		return 0;
}

int main(int argc, char **argv)
{
	int i, treeNo;
	long realLength, realInfSites;
	long simLength, simInfSites, realPolySites, simPolySites;
	double f, prob, cloneIndex, a, b, c;
	double realScore, *simScores, meanScore;
	TTree *tree;
	FILE *seqFile;
	clock_t totalStart;
	double totalSecs;
	
#ifdef __MWERKS__
	argc=ccommand(&argv);
#endif

	SetSeed(time(NULL));
	totalStart = clock();

	ReadParams(argc, argv);

	if (printHelp) {
 		PrintTitle();
		PrintUsage();
		exit(0);
	}

	if (!quiet)
 		PrintTitle();

	ReadFileParams();
	
	numBases=numSites;		
			
	if (!hasAlignment) {
		fprintf(stderr, "Sequence alignment missing from input file\n");
		exit(0);
	}
	
	if (statsOnly) {
		WriteStats();
		exit(0);
	}
	
	AllocateMemory();	
	ReadFile();
	
	SetModel(model);
	
	numTaxa=-1;
	
	if ((tree=NewTree())==NULL) {
		fprintf(stderr, "Out of memory\n");
		exit(0);
	}
			
	numTrees=OpenTreeFile();
	
	treeNo=0;

	ReadTree(tree_fv, tree, treeNo+1);
	
	numTaxa=tree->numTips;
	CreateSequences(tree);
	
	SetTipSequences(tree);

	realLength = CalcTreeLength(tree);
	realInfSites = GetNumInformativeSitesOnTree(tree, &realPolySites, twoStates);
	realScore = ((double)realInfSites) / realPolySites;
	
	if (verbose && !quiet) {
		PrintVerbose();
		InitProgressBar(numDatasets);
		DrawProgressBar();
	}
	
	if (verbose && !quiet)
		fprintf(stdout, "SimNo\tTreeLength\tPolymorphicSites\tInformSites\tPropInfSites\n");
		
		
	if (writeSeqs) {
		seqFile = fopen("pist.seq", "wt");
		if (seqFile == NULL) {
			fprintf(stderr, "Unable to open file to write sequences to: 'pist.seq'\n");
			exit(0);
		}
	}
	
	simScores = (double *)AllocMem(sizeof(double)*numDatasets, "simScores", "main", 0);
	meanScore = 0.0;
	
	i=0;
	while (i < numDatasets) {
		SetCategories();
		EvolveSequences(tree, -1.0, NULL);
		
		simLength = CalcTreeLength(tree);
		simInfSites = GetNumInformativeSitesOnTree(tree, &simPolySites, twoStates);
		
		f = 100.0 * fabs(simLength - realLength) / realLength;
		if ( f <= acceptInterval ) {
			simScores[i] = ((double) simInfSites) / simPolySites;
			meanScore += simScores[i];
			
			if (verbose && !quiet)
				fprintf(stdout, "%ld\t%ld\t%ld\t%ld\t%lf\n", i+1, simLength, simPolySites, simInfSites, simScores[i]);

			if (writeSeqs)
				WriteSequences(seqFile, tree);
				
			if (verbose && !quiet)
				ProgressBar();
				
			i++;
		}
	}
	meanScore /= numDatasets;
	
	if (writeSeqs)
		fclose(seqFile);

	qsort(simScores, numDatasets, sizeof(double), (void *)Compare);
	
	prob = 0.0;
	i = 0;
	while (i < numDatasets && realScore > simScores[i]) {
		prob++;
		i++;
	}
	prob = 1.0 - ((prob - 1.0) / numDatasets);

	if (!quiet) {
		fprintf(stderr, "\nInput tree has length %ld, with %ld polymorphic sites and %ld informative sites\n", realLength, realPolySites, realInfSites);
		fprintf(stderr, "\tScore = %lf\tProb(score) < %lf\n\n", realScore, prob);
	}
	
	if (!quiet && hasRandomLength) {
		a = (realScore - meanScore) / realLength;
		b = (1.0 - meanScore) / realLength;
		c = (1.0 - realScore) / randomLength;
		
		cloneIndex = a / (b - c);
		fprintf(stderr, "Informative Sites Index = %lf\n", cloneIndex);
	}
	
	DisposeTree(tree);
	
	FreeTree(tree);
	
	if (treeFile)
		fclose(tree_fv);

	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
	if (!quiet) {
		fprintf(stderr, "Time taken: %G seconds\n", totalSecs);
		if (verboseMemory)
			fprintf(stderr, "Total memory used: %ld\n", totalMem);
	}
	
	return 0;
}
